@@ -15,6 +15,7 @@
 #include <glib-unix.h>
 #include <locale.h>
 #include <polkit/polkit.h>
+#include <stdlib.h>
 #include <stdio.h>
 
 #include "power-profiles-daemon-resources.h"
@@ -36,15 +37,37 @@
 
 #define POWER_PROFILES_RESOURCES_PATH "/org/freedesktop/UPower/PowerProfiles"
 
+#define UPOWER_DBUS_NAME                  "org.freedesktop.UPower"
+#define UPOWER_DBUS_PATH                  "/org/freedesktop/UPower"
+#define UPOWER_DBUS_INTERFACE             "org.freedesktop.UPower"
+
+#define UPOWER_DBUS_DISPLAY_DEVICE_PATH   "/org/freedesktop/UPower/devices/DisplayDevice"
+#define UPOWER_DBUS_DEVICE_INTERFACE      "org.freedesktop.UPower.Device"
+
+#define LOGIND_DBUS_NAME                  "org.freedesktop.login1"
+#define LOGIND_DBUS_PATH                  "/org/freedesktop/login1"
+#define LOGIND_DBUS_INTERFACE             "org.freedesktop.login1.Manager"
+
 #ifndef POLKIT_HAS_AUTOPOINTERS
 /* FIXME: Remove this once we're fine to depend on polkit 0.114 */
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (PolkitAuthorizationResult, g_object_unref)
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (PolkitSubject, g_object_unref)
 #endif
 
+typedef struct {
+  GOptionGroup *group;
+  GLogLevelFlags log_level;
+  gboolean replace;
+  gboolean disable_upower;
+  gboolean disable_logind;
+  GStrv blocked_drivers;
+  GStrv blocked_actions;
+} DebugOptions;
+
 typedef struct {
   GMainLoop *main_loop;
   GDBusConnection *connection;
+  GCancellable *cancellable;
   guint name_id;
   guint legacy_name_id;
   gboolean was_started;
@@ -62,7 +85,18 @@ typedef struct {
   PpdDriverPlatform *platform_driver;
   GPtrArray *actions;
   GHashTable *profile_holds;
-  GLogLevelFlags log_level;
+
+  GDBusProxy *upower_proxy;
+  GDBusProxy *upower_display_proxy;
+  gulong upower_watch_id;
+  gulong upower_display_watch_id;
+  gulong upower_properties_id;
+  gulong upower_display_properties_id;
+  PpdPowerChangedReason power_changed_reason;
+
+  guint logind_sleep_signal_id;
+
+  DebugOptions *debug_options;
 } PpdApp;
 
 typedef struct {
@@ -73,6 +107,15 @@ typedef struct {
   char *requester_iface;
 } ProfileHold;
 
+static void
+debug_options_free(DebugOptions *options)
+{
+  g_strfreev (options->blocked_drivers);
+  g_strfreev (options->blocked_actions);
+  g_free (options);
+}
+G_DEFINE_AUTOPTR_CLEANUP_FUNC (DebugOptions, debug_options_free)
+
 static void
 profile_hold_free (ProfileHold *hold)
 {
@@ -89,10 +132,12 @@ static PpdApp *ppd_app = NULL;
 
 static void stop_profile_drivers (PpdApp *data);
 static void start_profile_drivers (PpdApp *data);
+static void upower_battery_set_power_changed_reason (PpdApp *, PpdPowerChangedReason);
 
 /* profile drivers and actions */
 #include "ppd-action-trickle-charge.h"
 #include "ppd-action-amdgpu-panel-power.h"
+#include "ppd-action-amdgpu-dpm.h"
 #include "ppd-driver-placeholder.h"
 #include "ppd-driver-platform-profile.h"
 #include "ppd-driver-intel-pstate.h"
@@ -114,6 +159,7 @@ static GTypeGetFunc objects[] = {
   /* Actions */
   ppd_action_trickle_charge_get_type,
   ppd_action_amdgpu_panel_power_get_type,
+  ppd_action_amdgpu_dpm_get_type,
 };
 
 typedef enum {
@@ -284,12 +330,12 @@ send_dbus_event_iface (PpdApp         *data,
   GVariantBuilder props_builder;
   GVariant *props_changed = NULL;
 
-  g_assert (data->connection);
+  g_return_if_fail (data->connection);
 
   if (mask == 0)
     return;
 
-  g_assert ((mask & PROP_ALL) != 0);
+  g_return_if_fail ((mask & PROP_ALL) != 0);
 
   g_variant_builder_init (&props_builder, G_VARIANT_TYPE ("a{sv}"));
 
@@ -429,12 +475,10 @@ actions_activate_profile (GPtrArray *actions,
   for (i = 0; i < actions->len; i++) {
     g_autoptr(GError) error = NULL;
     PpdAction *action;
-    gboolean ret;
 
     action = g_ptr_array_index (actions, i);
 
-    ret = ppd_action_activate_profile (action, profile, &error);
-    if (!ret)
+    if (!ppd_action_activate_profile (action, profile, &error))
       g_warning ("Failed to activate action '%s' to profile %s: %s",
                  ppd_profile_to_str (profile),
                  ppd_action_get_action_name (action),
@@ -449,52 +493,43 @@ activate_target_profile (PpdApp                      *data,
                          GError                     **error)
 {
   PpdProfile current_profile = data->active_profile;
-  gboolean cpu_set = TRUE;
-  gboolean platform_set = TRUE;
 
-  g_debug ("Setting active profile '%s' for reason '%s' (current: '%s')",
+  g_info ("Setting active profile '%s' for reason '%s' (current: '%s')",
            ppd_profile_to_str (target_profile),
            ppd_profile_activation_reason_to_str (reason),
            ppd_profile_to_str (current_profile));
 
   /* Try CPU first */
-  if (driver_profile_support (PPD_DRIVER (data->cpu_driver), target_profile))
-    cpu_set = ppd_driver_activate_profile (PPD_DRIVER (data->cpu_driver),
-                                           target_profile, reason, error);
-
-  if (!cpu_set) {
+  if (driver_profile_support (PPD_DRIVER (data->cpu_driver), target_profile) &&
+      !ppd_driver_activate_profile (PPD_DRIVER (data->cpu_driver),
+                                    target_profile, reason, error)) {
     g_prefix_error (error, "Failed to activate CPU driver '%s': ",
                     ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)));
     return FALSE;
   }
 
   /* Then try platform */
-  if (driver_profile_support (PPD_DRIVER (data->platform_driver), target_profile)) {
-    platform_set = ppd_driver_activate_profile (PPD_DRIVER (data->platform_driver),
-                                                target_profile, reason, error);
-  }
+  if (driver_profile_support (PPD_DRIVER (data->platform_driver), target_profile) &&
+      !ppd_driver_activate_profile (PPD_DRIVER (data->platform_driver),
+                                                target_profile, reason, error)) {
+    g_autoptr(GError) recovery_error = NULL;
 
-  if (!platform_set) {
     g_prefix_error (error, "Failed to activate platform driver '%s': ",
                     ppd_driver_get_driver_name (PPD_DRIVER (data->platform_driver)));
 
-    /* Try to recover */
-    if (cpu_set) {
-      g_autoptr(GError) recovery_error = NULL;
+    if (!PPD_IS_DRIVER (data->cpu_driver))
+      return FALSE;
 
-      g_debug ("Reverting CPU driver '%s' to profile '%s'",
-               ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
-               ppd_profile_to_str (current_profile));
+    g_debug ("Reverting CPU driver '%s' to profile '%s'",
+              ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
+              ppd_profile_to_str (current_profile));
 
-      if (!ppd_driver_activate_profile (PPD_DRIVER (data->cpu_driver),
-                                        current_profile, PPD_PROFILE_ACTIVATION_REASON_INTERNAL,
-                                        &recovery_error)) {
-        g_prefix_error (error, "Failed to revert CPU driver '%s': ",
-                        ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)));
-        g_warning ("Failed to revert CPU driver '%s': %s",
-                   ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
-                   recovery_error->message);
-      }
+    if (!ppd_driver_activate_profile (PPD_DRIVER (data->cpu_driver),
+                                      current_profile, PPD_PROFILE_ACTIVATION_REASON_INTERNAL,
+                                      &recovery_error)) {
+      g_warning ("Failed to revert CPU driver '%s': %s",
+                  ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
+                  recovery_error->message);
     }
 
     return FALSE;
@@ -827,7 +862,7 @@ handle_get_property (GDBusConnection *connection,
 {
   PpdApp *data = user_data;
 
-  g_assert (data->connection);
+  g_return_val_if_fail (data->connection, NULL);
 
   if (g_strcmp0 (property_name, "ActiveProfile") == 0)
     return g_variant_new_string (get_active_profile (data));
@@ -861,7 +896,7 @@ handle_set_property (GDBusConnection  *connection,
   PpdApp *data = user_data;
   const char *profile;
 
-  g_assert (data->connection);
+  g_return_val_if_fail (data->connection, FALSE);
 
   if (g_strcmp0 (property_name, "ActiveProfile") != 0) {
     g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
@@ -888,7 +923,7 @@ handle_method_call (GDBusConnection       *connection,
                     gpointer               user_data)
 {
   PpdApp *data = user_data;
-  g_assert (data->connection);
+  g_return_if_fail (data->connection);
 
   if (!g_str_equal (interface_name, POWER_PROFILES_IFACE_NAME) &&
       !g_str_equal (interface_name, POWER_PROFILES_LEGACY_IFACE_NAME)) {
@@ -947,9 +982,9 @@ name_lost_handler (GDBusConnection *connection,
   PpdBusOwnData *data = user_data;
   PpdApp *app = data->app;
 
-  g_debug ("power-profiles-daemon is already running, or it cannot own its D-Bus name. Verify installation.");
+  g_warning ("power-profiles-daemon is already running, or it cannot own its D-Bus name. Verify installation.");
   if (!app->was_started)
-    app->ret = 1;
+    app->ret = EXIT_FAILURE;
 
   g_main_loop_quit (app->main_loop);
 }
@@ -996,6 +1031,286 @@ bus_acquired_handler (GDBusConnection *connection,
   data->app->connection = g_object_ref (connection);
 }
 
+static void
+upower_source_update_from_value (PpdApp   *data,
+                                 GVariant *battery_val)
+{
+  PpdPowerChangedReason reason;
+
+  if (!battery_val)
+    reason = PPD_POWER_CHANGED_REASON_UNKNOWN;
+  else if (g_variant_get_boolean (battery_val))
+    reason = PPD_POWER_CHANGED_REASON_BATTERY;
+  else
+    reason = PPD_POWER_CHANGED_REASON_AC;
+
+  upower_battery_set_power_changed_reason (data, reason);
+}
+
+static void
+upower_source_update (PpdApp *data)
+{
+  g_autoptr(GVariant) battery_val = NULL;
+
+  battery_val = g_dbus_proxy_get_cached_property (data->upower_proxy, "OnBattery");
+  upower_source_update_from_value (data, battery_val);
+}
+
+static void
+upower_battery_changed (PpdApp *data, gdouble level)
+{
+  g_info ("Battery level changed to %f", level);
+
+  for (guint i = 0; i < data->actions->len; i++) {
+    g_autoptr(GError) error = NULL;
+    PpdAction *action;
+
+    action = g_ptr_array_index (data->actions, i);
+
+    if (!ppd_action_battery_changed (action, level, &error)) {
+      g_warning ("failed to update action %s: %s",
+                 ppd_action_get_action_name (action),
+                 error->message);
+      g_clear_error (&error);
+      continue;
+    }
+  }
+
+  if (PPD_IS_DRIVER_CPU (data->cpu_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_battery_changed (PPD_DRIVER (data->cpu_driver), level, &error)) {
+      g_warning ("failed to update driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
+                 error->message);
+    }
+  }
+
+  if (PPD_IS_DRIVER_PLATFORM (data->platform_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_battery_changed (PPD_DRIVER (data->platform_driver), level, &error)) {
+      g_warning ("failed to update driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->platform_driver)),
+                 error->message);
+    }
+  }
+}
+
+static void
+upower_battery_update (PpdApp *data)
+{
+  g_autoptr(GVariant) val = NULL;
+
+  val = g_dbus_proxy_get_cached_property (data->upower_display_proxy, "Percentage");
+  if (val)
+    upower_battery_changed(data, g_variant_get_double (val));
+}
+
+static void
+upower_properties_changed (GDBusProxy *proxy,
+                           GVariant *changed_properties,
+                           GStrv invalidated_properties,
+                           PpdApp *data)
+{
+  g_auto(GVariantDict) props_dict = G_VARIANT_DICT_INIT (changed_properties);
+  g_autoptr(GVariant) battery_val = NULL;
+  g_autoptr(GVariant) percent_val = NULL;
+
+  battery_val = g_variant_dict_lookup_value (&props_dict, "OnBattery",
+                                             G_VARIANT_TYPE_BOOLEAN);
+
+  if (battery_val)
+    upower_source_update_from_value (data, battery_val);
+
+  percent_val = g_variant_dict_lookup_value (&props_dict, "Percentage",
+                                             G_VARIANT_TYPE_DOUBLE);
+  if (percent_val)
+    upower_battery_changed (data, g_variant_get_double (percent_val));
+}
+
+static void
+upower_battery_set_power_changed_reason (PpdApp                *data,
+                                         PpdPowerChangedReason  reason)
+{
+  if (data->power_changed_reason == reason)
+    return;
+
+  data->power_changed_reason = reason;
+  g_info ("Power Changed because of reason %s",
+          ppd_power_changed_reason_to_str (reason));
+
+  for (guint i = 0; i < data->actions->len; i++) {
+    g_autoptr(GError) error = NULL;
+    PpdAction *action;
+
+    action = g_ptr_array_index (data->actions, i);
+
+    if (!ppd_action_power_changed (action, reason, &error)) {
+      g_warning ("failed to update action %s: %s",
+                 ppd_action_get_action_name (action),
+                 error->message);
+      g_clear_error (&error);
+      continue;
+    }
+  }
+
+  if (PPD_IS_DRIVER_CPU (data->cpu_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_power_changed (PPD_DRIVER (data->cpu_driver), reason, &error)) {
+      g_warning ("failed to update driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
+                 error->message);
+    }
+  }
+
+  if (PPD_IS_DRIVER_PLATFORM (data->platform_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_power_changed (PPD_DRIVER (data->platform_driver), reason, &error)) {
+      g_warning ("failed to update driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->platform_driver)),
+                 error->message);
+    }
+  }
+}
+
+static void
+upower_name_owner_changed (GObject    *object,
+                           GParamSpec *pspec,
+                           gpointer    user_data)
+{
+  PpdApp *data = user_data;
+  GDBusProxy *upower_proxy = G_DBUS_PROXY (object);
+  g_autofree char *name_owner = NULL;
+
+  name_owner = g_dbus_proxy_get_name_owner (upower_proxy);
+
+  if (name_owner != NULL) {
+    g_debug ("%s appeared", UPOWER_DBUS_NAME);
+    upower_source_update (data);
+    return;
+  }
+
+  g_debug ("%s vanished", UPOWER_DBUS_NAME);
+  upower_battery_set_power_changed_reason (data, PPD_POWER_CHANGED_REASON_UNKNOWN);
+}
+
+static void
+on_upower_proxy_cb (GObject *source_object,
+                    GAsyncResult *res,
+                    gpointer user_data)
+{
+  PpdApp *data = user_data;
+  g_autoptr(GDBusProxy) upower_proxy = NULL;
+  g_autoptr(GError) error = NULL;
+
+  upower_proxy = g_dbus_proxy_new_finish (res, &error);
+
+  if (upower_proxy == NULL) {
+    if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+      return;
+
+    g_warning ("failed to connect to upower: %s", error->message);
+    return;
+  }
+
+  g_return_if_fail (data->upower_proxy == NULL);
+  data->upower_proxy = g_steal_pointer (&upower_proxy);
+
+  data->upower_properties_id = g_signal_connect (data->upower_proxy,
+                                                 "g-properties-changed",
+                                                 G_CALLBACK (upower_properties_changed),
+                                                 data);
+
+  if (!data->upower_display_watch_id) {
+    data->upower_watch_id = g_signal_connect (data->upower_proxy,
+                                              "notify::g-name-owner",
+                                              G_CALLBACK (upower_name_owner_changed),
+                                              data);
+  }
+
+  upower_source_update (data);
+}
+
+static void
+on_upower_display_proxy_cb (GObject *source_object,
+                            GAsyncResult *res,
+                            gpointer user_data)
+{
+  PpdApp *data = user_data;
+  g_autoptr(GDBusProxy) proxy = NULL;
+  g_autoptr(GError) error = NULL;
+
+  proxy = g_dbus_proxy_new_finish (res, &error);
+
+  if (proxy == NULL) {
+    if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+      return;
+
+    g_warning ("failed to connect to upower: %s", error->message);
+    return;
+  }
+
+  g_return_if_fail (data->upower_display_proxy == NULL);
+  data->upower_display_proxy = g_steal_pointer (&proxy);
+
+  data->upower_display_properties_id = g_signal_connect (data->upower_display_proxy,
+                                                         "g-properties-changed",
+                                                         G_CALLBACK (upower_properties_changed),
+                                                         data);
+
+  if (!data->upower_watch_id) {
+    data->upower_display_watch_id = g_signal_connect (data->upower_display_proxy,
+                                                      "notify::g-name-owner",
+                                                      G_CALLBACK (upower_name_owner_changed),
+                                                      data);
+  }
+
+  upower_battery_update (data);
+}
+
+static void
+on_logind_prepare_for_sleep_cb (GDBusConnection *connection,
+                                const gchar     *sender_name,
+                                const gchar     *object_path,
+                                const gchar     *interface_name,
+                                const gchar     *signal_name,
+                                GVariant        *parameters,
+                                gpointer         user_data)
+{
+  PpdApp *data = user_data;
+  gboolean start;
+
+  g_variant_get (parameters, "(b)", &start);
+
+  if (start)
+    g_debug ("System preparing for suspend");
+  else
+    g_debug ("System woke up from suspend");
+
+  if (PPD_IS_DRIVER_CPU (data->cpu_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_prepare_to_sleep (PPD_DRIVER (data->cpu_driver), start, &error)) {
+      g_warning ("failed to notify driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->cpu_driver)),
+                 error->message);
+    }
+  }
+
+  if (PPD_IS_DRIVER_PLATFORM (data->platform_driver)) {
+    g_autoptr(GError) error = NULL;
+
+    if (!ppd_driver_prepare_to_sleep (PPD_DRIVER (data->platform_driver), start, &error)) {
+      g_warning ("failed to notify driver %s: %s",
+                 ppd_driver_get_driver_name (PPD_DRIVER (data->platform_driver)),
+                 error->message);
+    }
+  }
+}
+
 static gboolean
 has_required_drivers (PpdApp *data)
 {
@@ -1010,52 +1325,101 @@ has_required_drivers (PpdApp *data)
 }
 
 static void
-driver_probe_request_cb (PpdDriver *driver,
-                         gpointer   user_data)
+restart_profile_drivers (PpdApp *data)
 {
-  PpdApp *data = user_data;
-
   stop_profile_drivers (data);
   start_profile_drivers (data);
 }
 
+static void
+driver_probe_request_cb (PpdDriver *driver,
+                         gpointer   user_data)
+{
+  PpdApp *data = user_data;
+
+  restart_profile_drivers (data);
+}
+
+static void
+disconnect_array_objects_signals_by_data (GPtrArray *array,
+                                          void      *data)
+{
+  for (guint i = 0; i < array->len; ++i) {
+    GObject *object = g_ptr_array_index (array, i);
+
+    g_signal_handlers_disconnect_by_data (object, data);
+  }
+}
+
+static void
+maybe_disconnect_object_by_data (void *object,
+                                 void *data)
+{
+  if (!G_IS_OBJECT (object))
+    return;
+
+  g_signal_handlers_disconnect_by_data (object, data);
+}
+
 static void
 stop_profile_drivers (PpdApp *data)
 {
+  if (data->logind_sleep_signal_id) {
+    g_dbus_connection_signal_unsubscribe (data->connection, data->logind_sleep_signal_id);
+    data->logind_sleep_signal_id = 0;
+  }
+
+  upower_battery_set_power_changed_reason (data, PPD_POWER_CHANGED_REASON_UNKNOWN);
   release_all_profile_holds (data);
+  g_cancellable_cancel (data->cancellable);
+  disconnect_array_objects_signals_by_data (data->probed_drivers, data);
   g_ptr_array_set_size (data->probed_drivers, 0);
+  disconnect_array_objects_signals_by_data (data->actions, data);
   g_ptr_array_set_size (data->actions, 0);
+  g_clear_signal_handler (&data->upower_watch_id, data->upower_proxy);
+  g_clear_signal_handler (&data->upower_properties_id, data->upower_proxy);
+  g_clear_signal_handler (&data->upower_display_watch_id, data->upower_display_proxy);
+  g_clear_signal_handler (&data->upower_display_properties_id, data->upower_display_proxy);
+  g_clear_object (&data->cancellable);
+  maybe_disconnect_object_by_data (data->upower_proxy, data);
+  g_clear_object (&data->upower_proxy);
+  maybe_disconnect_object_by_data (data->upower_display_proxy, data);
+  g_clear_object (&data->upower_display_proxy);
+  maybe_disconnect_object_by_data (data->cpu_driver, data);
   g_clear_object (&data->cpu_driver);
+  maybe_disconnect_object_by_data (data->platform_driver, data);
   g_clear_object (&data->platform_driver);
 }
 
 static gboolean
-action_blocked (PpdAction *action)
+action_blocked (PpdApp *app, PpdAction *action)
 {
   const gchar *action_name = ppd_action_get_action_name (action);
-  const gchar *env = g_getenv ("POWER_PROFILE_DAEMON_ACTION_BLOCK");
+  gboolean blocked;
 
-  if (env == NULL)
+  if (app->debug_options->blocked_actions == NULL || g_strv_length (app->debug_options->blocked_actions) == 0)
     return FALSE;
 
-  g_auto(GStrv) actions = NULL;
+  blocked = g_strv_contains ((const gchar *const *) app->debug_options->blocked_actions, action_name);
 
-  actions = g_strsplit (env, ",", -1);
-  return g_strv_contains ((const gchar *const *)actions, action_name);
+  if (blocked)
+    g_debug ("Action '%s' is blocked", action_name);
+  return blocked;
 }
 
 static gboolean
-driver_blocked (PpdDriver *driver)
+driver_blocked (PpdApp *app, PpdDriver *driver)
 {
   const gchar *driver_name = ppd_driver_get_driver_name (driver);
-  const gchar *env = g_getenv ("POWER_PROFILE_DAEMON_DRIVER_BLOCK");
+  gboolean blocked;
 
-  if (env == NULL)
+  if (app->debug_options->blocked_drivers == NULL || g_strv_length (app->debug_options->blocked_drivers) == 0)
     return FALSE;
 
-  g_auto(GStrv) drivers = NULL;
-  drivers = g_strsplit (env, ",", -1);
-  return g_strv_contains ((const char **) drivers, driver_name);
+  blocked = g_strv_contains ((const gchar *const *) app->debug_options->blocked_drivers, driver_name);
+  if (blocked)
+    g_debug ("Driver '%s' is blocked", driver_name);
+  return blocked;
 }
 
 static void
@@ -1063,6 +1427,11 @@ start_profile_drivers (PpdApp *data)
 {
   guint i;
   g_autoptr(GError) initial_error = NULL;
+  gboolean needs_battery_state_monitor = FALSE;
+  gboolean needs_battery_change_monitor = FALSE;
+  gboolean needs_suspend_monitor = FALSE;
+
+  data->cancellable = g_cancellable_new ();
 
   for (i = 0; i < G_N_ELEMENTS (objects); i++) {
     g_autoptr(GObject) object = NULL;
@@ -1075,7 +1444,7 @@ start_profile_drivers (PpdApp *data)
       PpdProbeResult result;
 
       g_debug ("Handling driver '%s'", ppd_driver_get_driver_name (driver));
-      if (driver_blocked (driver)) {
+      if (driver_blocked (data, driver)) {
         g_debug ("Driver '%s' is blocked, skipping", ppd_driver_get_driver_name (driver));
         continue;
       }
@@ -1121,7 +1490,18 @@ start_profile_drivers (PpdApp *data)
       else if (PPD_IS_DRIVER_PLATFORM (driver))
         g_set_object (&data->platform_driver, PPD_DRIVER_PLATFORM (driver));
       else
-        g_assert_not_reached ();
+        g_return_if_reached ();
+
+      if (PPD_DRIVER_GET_CLASS (driver)->power_changed != NULL)
+        needs_battery_state_monitor = TRUE;
+
+      if (PPD_DRIVER_GET_CLASS (driver)->battery_changed != NULL)
+        needs_battery_change_monitor = TRUE;
+
+      if (PPD_DRIVER_GET_CLASS (driver)->prepare_to_sleep != NULL)
+        needs_suspend_monitor = TRUE;
+
+      g_info ("Driver '%s' loaded", ppd_driver_get_driver_name (driver));
 
       g_signal_connect (G_OBJECT (driver), "notify::performance-degraded",
                         G_CALLBACK (driver_performance_degraded_changed_cb), data);
@@ -1135,7 +1515,7 @@ start_profile_drivers (PpdApp *data)
 
       g_debug ("Handling action '%s'", ppd_action_get_action_name (action));
 
-      if (action_blocked (action)) {
+      if (action_blocked (data, action)) {
         g_debug ("Action '%s' is blocked, skipping", ppd_action_get_action_name (action));
         continue;
       }
@@ -1146,16 +1526,24 @@ start_profile_drivers (PpdApp *data)
         continue;
       }
 
+      if (PPD_ACTION_GET_CLASS (action)->power_changed != NULL)
+        needs_battery_state_monitor = TRUE;
+
+      if (PPD_ACTION_GET_CLASS (action)->battery_changed != NULL)
+        needs_battery_change_monitor = TRUE;
+
+      g_info ("Action '%s' loaded", ppd_action_get_action_name (action));
       g_ptr_array_add (data->actions, g_steal_pointer (&action));
       continue;
     }
 
-    g_assert_not_reached ();
+    g_return_if_reached ();
   }
 
   if (!has_required_drivers (data)) {
+    data->ret = EXIT_FAILURE;
     g_warning ("Some non-optional profile drivers are missing, programmer error");
-    goto bail;
+    g_main_loop_quit (data->main_loop);
   }
 
   /* Set initial state either from configuration, or using the currently selected profile */
@@ -1164,24 +1552,70 @@ start_profile_drivers (PpdApp *data)
     g_warning ("Failed to activate initial profile: %s", initial_error->message);
 
   send_dbus_event (data, PROP_ALL);
-
   data->was_started = TRUE;
 
-  return;
+  if (data->debug_options->disable_upower) {
+    g_debug ("upower is disabled, let's skip it");
+  } else if (needs_battery_state_monitor || needs_battery_change_monitor) {
+    /* start watching for power changes */
+    if (needs_battery_state_monitor) {
+      g_debug ("Battery state monitor required, connecting to upower...");
+      g_dbus_proxy_new (data->connection,
+                        G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START |
+                        G_DBUS_PROXY_FLAGS_DO_NOT_CONNECT_SIGNALS,
+                        NULL,
+                        UPOWER_DBUS_NAME,
+                        UPOWER_DBUS_PATH,
+                        UPOWER_DBUS_INTERFACE,
+                        data->cancellable,
+                        on_upower_proxy_cb,
+                        data);
+    }
 
-bail:
-  data->ret = 1;
-  g_debug ("Exiting because some non recoverable error occurred during startup");
-  g_main_loop_quit (data->main_loop);
+    if (needs_battery_change_monitor) {
+      g_debug ("Battery change monitor required, connecting to upower...");
+      g_dbus_proxy_new (data->connection,
+                        G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START |
+                        G_DBUS_PROXY_FLAGS_DO_NOT_CONNECT_SIGNALS,
+                        NULL,
+                        UPOWER_DBUS_NAME,
+                        UPOWER_DBUS_DISPLAY_DEVICE_PATH,
+                        UPOWER_DBUS_DEVICE_INTERFACE,
+                        data->cancellable,
+                        on_upower_display_proxy_cb,
+                        data);
+    }
+  } else {
+    g_debug ("No battery state monitor required by any driver, let's skip it");
+  }
+
+  if (data->debug_options->disable_logind) {
+    g_debug ("logind is disabled, let's skip it");
+  } else if (needs_suspend_monitor) {
+    g_debug ("Suspension state monitor required, monitoring logind...");
+    data->logind_sleep_signal_id =
+      g_dbus_connection_signal_subscribe (data->connection,
+                                          LOGIND_DBUS_NAME,
+                                          LOGIND_DBUS_INTERFACE,
+                                          "PrepareForSleep",
+                                          LOGIND_DBUS_PATH,
+                                          NULL,
+                                          G_DBUS_SIGNAL_FLAGS_NONE,
+                                          on_logind_prepare_for_sleep_cb,
+                                          data,
+                                          NULL);
+  } else {
+    g_debug ("No suspension monitor required by any driver, let's skip it");
+  }
 }
 
 void
-restart_profile_drivers (void)
+restart_profile_drivers_for_default_app (void)
 {
-  stop_profile_drivers (ppd_app);
-  start_profile_drivers (ppd_app);
+  restart_profile_drivers (ppd_app);
 }
 
+
 static void
 name_acquired_handler (GDBusConnection *connection,
                        const gchar     *name,
@@ -1196,7 +1630,6 @@ name_acquired_handler (GDBusConnection *connection,
 
 static gboolean
 setup_dbus (PpdApp    *data,
-            gboolean   replace,
             GError   **error)
 {
   g_autoptr(GBytes) iface_data = NULL;
@@ -1235,7 +1668,7 @@ setup_dbus (PpdApp    *data,
   own_data->legacy_interface = g_dbus_interface_info_ref (legacy_introspection_data->interfaces[0]);
 
   own_data->flags = G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT;
-  if (replace)
+  if (data->debug_options->replace)
     own_data->flags |= G_BUS_NAME_OWNER_FLAGS_REPLACE;
 
   data->name_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
@@ -1256,13 +1689,16 @@ free_app_data (PpdApp *data)
   if (data == NULL)
     return;
 
+  stop_profile_drivers (data);
+
   g_clear_handle_id (&data->name_id, g_bus_unown_name);
   g_clear_handle_id (&data->legacy_name_id, g_bus_unown_name);
 
+  g_clear_pointer (&data->debug_options, debug_options_free);
   g_clear_pointer (&data->config_path, g_free);
   g_clear_pointer (&data->config, g_key_file_unref);
-  g_ptr_array_free (data->probed_drivers, TRUE);
-  g_ptr_array_free (data->actions, TRUE);
+  g_clear_pointer (&data->probed_drivers, g_ptr_array_unref);
+  g_clear_pointer (&data->actions, g_ptr_array_unref);
   g_clear_object (&data->cpu_driver);
   g_clear_object (&data->platform_driver);
   g_hash_table_destroy (data->profile_holds);
@@ -1297,7 +1733,7 @@ debug_handler_cb (const gchar *log_domain,
                   const gchar *message,
                   gpointer user_data)
 {
-  PpdApp *data = user_data;
+  DebugOptions *data = user_data;
   g_autoptr(GString) domain = NULL;
   gboolean use_color;
   gint color;
@@ -1339,22 +1775,132 @@ quit_signal_callback (gpointer user_data)
   return FALSE;
 }
 
+static gboolean
+verbose_arg_cb (const gchar *option_name,
+                const gchar *value,
+                gpointer user_data,
+                GError **error)
+{
+  DebugOptions *data = user_data;
+
+  if (data->log_level == G_LOG_LEVEL_MESSAGE) {
+          data->log_level = G_LOG_LEVEL_INFO;
+          return TRUE;
+  }
+  if (data->log_level == G_LOG_LEVEL_INFO) {
+          data->log_level = G_LOG_LEVEL_DEBUG;
+          return TRUE;
+  }
+  g_set_error_literal (error,
+                       G_OPTION_ERROR,
+                       G_OPTION_ERROR_FAILED,
+                       "No further debug level supported");
+  return FALSE;
+}
+
+static gboolean
+debug_pre_parse_hook (GOptionContext *context,
+                      GOptionGroup *group,
+                      gpointer user_data,
+                      GError **error)
+{
+  DebugOptions *data = user_data;
+
+  const GOptionEntry options[] = {
+    {
+      "verbose",
+      'v',
+      G_OPTION_FLAG_NO_ARG,
+      G_OPTION_ARG_CALLBACK,
+      (GOptionArgFunc)verbose_arg_cb,
+      "Show extra debugging information",
+      NULL,
+    },
+    {
+      "replace",
+      'r',
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_NONE,
+      &data->replace,
+      "Replace the running instance of power-profiles-daemon",
+      NULL,
+    },
+    {
+      "block-driver",
+      0,
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_STRING_ARRAY,
+      &data->blocked_drivers,
+      "Block driver(s) from loading",
+      NULL,
+    },
+    {
+      "block-action",
+      0,
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_STRING_ARRAY,
+      &data->blocked_actions,
+      "Block action(s) from loading",
+      NULL,
+    },
+    {
+      "disable-upower",
+      0,
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_NONE,
+      &data->disable_upower,
+      "Disable upower integration",
+      NULL,
+    },
+    {
+      "disable-logind",
+      0,
+      G_OPTION_FLAG_NONE,
+      G_OPTION_ARG_NONE,
+      &data->disable_logind,
+      "Disable logind integration",
+      NULL,
+    },
+    { NULL }
+  };
+  g_option_group_add_entries (group, options);
+
+  return TRUE;
+}
+
+static gboolean
+debug_post_parse_hook (GOptionContext *context,
+                       GOptionGroup *group,
+                       gpointer user_data,
+                       GError **error)
+{
+  DebugOptions *debug = (DebugOptions *)user_data;
+
+  g_log_set_default_handler (debug_handler_cb, debug);
+
+  return TRUE;
+}
+
 int main (int argc, char **argv)
 {
-  g_autoptr (PpdApp) data = NULL;
+  g_autoptr(DebugOptions) debug_options = g_new0 (DebugOptions, 1);
+  g_autoptr(PpdApp) data = NULL;
   g_autoptr(GOptionContext) option_context = NULL;
   g_autoptr(GError) error = NULL;
-  gboolean verbose = FALSE;
-  gboolean replace = FALSE;
-  const GOptionEntry options[] = {
-    { "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose, "Show extra debugging information", NULL },
-    { "replace", 'r', 0, G_OPTION_ARG_NONE, &replace, "Replace the running instance of power-profiles-daemon", NULL },
-    { NULL}
-  };
+
+  debug_options->log_level = G_LOG_LEVEL_MESSAGE;
+  debug_options->group = g_option_group_new ("debug",
+                                             "Debugging Options",
+                                             "Show debugging options",
+                                             debug_options,
+                                             NULL);
+  g_option_group_set_parse_hooks (debug_options->group,
+                                  debug_pre_parse_hook,
+                                  debug_post_parse_hook);
 
   setlocale (LC_ALL, "");
   option_context = g_option_context_new ("");
-  g_option_context_add_main_entries (option_context, options, NULL);
+  g_option_context_add_group (option_context, debug_options->group);
 
   if (!g_option_context_parse (option_context, &argc, &argv, &error)) {
     g_print ("Failed to parse arguments: %s\n", error->message);
@@ -1369,24 +1915,20 @@ int main (int argc, char **argv)
   data->profile_holds = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) profile_hold_free);
   data->active_profile = PPD_PROFILE_BALANCED;
   data->selected_profile = PPD_PROFILE_BALANCED;
-  data->log_level = verbose ? G_LOG_LEVEL_DEBUG : G_LOG_LEVEL_MESSAGE;
+  data->debug_options = g_steal_pointer(&debug_options);
 
   g_unix_signal_add (SIGTERM, quit_signal_callback, data);
   g_unix_signal_add (SIGINT, quit_signal_callback, data);
 
-  /* redirect all domains */
-  if (verbose && !g_log_writer_is_journald (fileno (stdout)))
-    g_log_set_default_handler (debug_handler_cb, data);
-
-  g_debug ("Starting power-profiles-daemon version "VERSION);
+  g_info ("Starting power-profiles-daemon version "VERSION);
 
   load_configuration (data);
   ppd_app = data;
 
   /* Set up D-Bus */
-  if (!setup_dbus (data, replace, &error)) {
+  if (!setup_dbus (data, &error)) {
     g_error ("Failed to start dbus: %s", error->message);
-    return 1;
+    return EXIT_FAILURE;
   }
 
   g_main_loop_run (data->main_loop);
